package visl.rasmus.gui;

/*
 * @(#)ScrollablePanel.java     0.1 96/10/29 Kul Bhatt
 *
 * Copyright (c) 1996 Kul Bhatt
 *
 *  This program is distributed freely. You can copy it, use it, and
 *  distribute it at your free will.
 *  Author provides no WARRANTY; Not even the implied warranty of USABILITY
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  Author is NOT responsible for any damages caused by direct or
 *  indirect use of this program. 
 *  
 *  k.bhatt@ieee.org 
 */

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Panel;
import java.awt.Scrollbar;

/**
 *  ScrollablePanel - is a scrollable Panel!
 *  It allows you to scroll any arbitrary component including images
 *  and other containers.
 *
 *  It creates two scroll bars around the component to be scrolled.
 *
 *<PRE>
 *       ---------
 *       |    ||
 *       |    ||
 *       | Comp  ||Scrollbar
 *       |    ||
 *       |    ||
 *       |    ||
 *       _________
 *       ---------
 *       Scrollbar
 *</PRE>
 *  The ScrollablePanel is resizable.
 *
 *  How to use ScrollablePanel?
 *  Create a Scrollable Panel with
 *       new ScrollablePanel(Component scrolledComponent)
 *  add this to your container as usual.
 *@author Kul Bhatt
 *@author Martin Carlsen (slight modifications to scroll, layoutContainer, manageScrollbars)
 * @author Rasmus Lock Larsen (addded a cursor + handling)
 */

public class ScrollablePanel extends Panel implements LayoutManager {
    Panel viewPort;
    Scrollbar vbar;
    Scrollbar hbar;

    Component scrolledComponent;
 
    /**
     * ScrollablePanel(Component scrolledComponent)
     *   creates a Scrollable Panel that can scroll scrolledComponent
     */
    public ScrollablePanel(Component scrolledComponent)
    {
	this.scrolledComponent = scrolledComponent;
	viewPort = new Panel();
	hbar = new Scrollbar(Scrollbar.HORIZONTAL);
	vbar = new Scrollbar(Scrollbar.VERTICAL);
	setLayout(this);
	add("Not used", vbar);
	add("Not", hbar);
	add("Not", viewPort);
	viewPort.setLayout(new LazyLayout());
	viewPort.add("scroll!", scrolledComponent);
    }
    public ScrollablePanel(Panel scrolledComponent)
    {
	this.scrolledComponent = scrolledComponent;
	viewPort = new Panel();
	hbar = new Scrollbar(Scrollbar.HORIZONTAL);
	vbar = new Scrollbar(Scrollbar.VERTICAL);
	setLayout(this);
	add("Not used", vbar);
	add("Not", hbar);
	add("Not", viewPort);
	viewPort.setLayout(new LazyLayout());
	viewPort.add("scroll!", scrolledComponent);
    }
    

    public void addLayoutComponent(String name, Component comp) {}
 
    public void layoutContainer(Container parent)
    {
	viewPort.validate();
	Insets insets = parent.insets();
	Dimension parentSize = parent.size();
	int top = insets.top;
	int left = insets.left;
	int bottom = parentSize.height - insets.bottom;
	int right = parentSize.width - insets.right;
  
	Dimension pv = vbar.preferredSize();
	Dimension ph = hbar.preferredSize();
  
	vbar.reshape(right - pv.width, 
		     top, 
		     pv.width, 
		     bottom - top - ph.height);
	hbar.reshape(left, 
		     bottom - ph.height, 
		     right - left - pv.width,
		     ph.height);
	viewPort.reshape(left, 
			 top, 
			 right - vbar.size().width, 
			 bottom - hbar.size().height);
	manageScrollbars();
	scroll();
    }
    
    public Dimension minimumLayoutSize(Container parent) {
	Dimension d = new Dimension(scrolledComponent.minimumSize());
	d.height += hbar.preferredSize().height;
	d.width += vbar.preferredSize().width;
	return d;  
    }
 
    public Dimension preferredLayoutSize(Container parent) {
	Dimension d = new Dimension(scrolledComponent.preferredSize());
	d.height += hbar.preferredSize().height;
	d.width += vbar.preferredSize().width;
	return d;
    }
    
    public void removeLayoutComponent(Component comp) { 
    }
    
    private final void manageScrollbars() {
	Dimension v = vbar.size();
	Dimension h = hbar.size();
	Dimension sc = scrolledComponent.size();
  
	vbar.setValues(vbar.getValue(), v.height, 0, sc.height/* - v.height*/);
	hbar.setValues(hbar.getValue(), h.width, 0, sc.width/* - h.width*/);
	hbar.setLineIncrement(sc.width/12);
	hbar.setPageIncrement(sc.width/4);
	vbar.setLineIncrement(sc.height/12);
	vbar.setPageIncrement(sc.height/4);
    }
 
    public Dimension preferredSize()
    {
	Dimension d = new Dimension(scrolledComponent.preferredSize());
	d.height += hbar.preferredSize().height;
	d.width += vbar.preferredSize().width;
	return d;
    }
 
    public Dimension minimumSize() {
	return preferredSize();
    }
    /** Handles the events
     * @author Rasmus Lock Larsen, VISL, october 2000 - added DRAG, MOVE */
    public boolean handleEvent(Event e) {
	switch (e.id){
	case Event.SCROLL_LINE_UP:
	case Event.SCROLL_LINE_DOWN:
	case Event.SCROLL_PAGE_UP:
	case Event.SCROLL_PAGE_DOWN:
	case Event.SCROLL_ABSOLUTE:
	    scroll(); break;
	case Event.MOUSE_DRAG:
	case Event.MOUSE_MOVE:
	    if (scrolledComponent.bounds().inside(e.x, e.y)) {
		int x = scrolledComponent.bounds().x;
		int y = scrolledComponent.bounds().y;
		e.translate(-x,-y);
		scrolledComponent.handleEvent(e);
		return true;
	    } else if (viewPort.bounds().inside(e.x, e.y)) {
		int x = viewPort.bounds().x;
		int y = viewPort.bounds().y;
		e.translate(-x,-y);
		viewPort.handleEvent(e);
		return true;
	    }
	    break;
	}
	return scrolledComponent.handleEvent(e);
    }
    private final void scroll() {
	Dimension pv = viewPort.size();
	Dimension ph = scrolledComponent.size();
	int moveX = (pv.width > ph.width) ? pv.width/2 - ph.width/2 : -hbar.getValue();
	int moveY = (pv.height > ph.height) ? 0 /*pv.height/2 - ph.height/2*/ : -vbar.getValue();;
	scrolledComponent.move(moveX, moveY);
    }
    public void repaint() {
	super.repaint();
	viewPort.repaint();
	scrolledComponent.repaint();
    }
}



