package visl.rasmus.hitserver.client;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.net.Socket;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

import visl.rasmus.hitserver.data.HighScoreItem;
import visl.rasmus.hitserver.data.Sentence;
import visl.rasmus.hitserver.data.TreeNode;
import visl.rasmus.hitserver.server.HitServerInterface;
import visl.rasmus.utils.Timer;
import visl.rasmus.utils.TimerListener;
import visl.resources.Dict;

/** This class implements the public interface to the HitServer 
 * @see HitXtraClient
 * @author Rasmus Lock Larsen, VISL March 2001 */
public class HitClient implements HitServerInterface, TimerListener {
	public static final Hashtable short2long = new Hashtable();
	/** Time Out set to 50 seconds! */
	public static final long TIME_TIMEOUT = 50000;

	protected String servername;
	protected Socket socket;
	protected DataInputStream in;
	protected PrintStream out;
	protected String language, game;
	protected Timer timer;
	protected long time, start_time;
	protected boolean timing, timeout;
	protected int port;

	public HitClient(
		String servername,
		int port,
		String language,
		String game) {
		this.servername = servername;
		this.language = language;
		this.game = game;
		this.port = port;
		if (!Dict.isInitialized())
			Dict.localize();
		// This should be removed - not suitable to have statically linked
		// "dictionaries".
		short2long.put(T100Y, Dict.get("This year's top 100 scores!"));
		short2long.put(T100M, Dict.get("This month's top 100 scores!"));
		short2long.put(T10W, Dict.get("This week's top 10 scores!"));
		short2long.put(
			T10D,
			Dict.get("The top 10 scores in the last 24 hours!"));
		short2long.put("en", "English");
		short2long.put("bs", "Bosnian");
		short2long.put("el", "Greek");
		short2long.put("it", "Italian");
		short2long.put("la", "Latin");
		short2long.put("pt", "Portuguese");
		short2long.put("da", "Danish");
		short2long.put("fr", "French");
		short2long.put("de", "German");

		// This should be placed elsewhere
		short2long.put("n", Dict.get("nouns"));
		short2long.put("v", Dict.get("verbs"));
		short2long.put("adj", Dict.get("adjectives"));
		short2long.put("adv", Dict.get("adverbs"));
		short2long.put("art", Dict.get("articles"));
		short2long.put("pron", Dict.get("pronouns"));
		short2long.put("prp", Dict.get("prepositions"));
		short2long.put("conj", Dict.get("conjunctions"));
		short2long.put("infm", Dict.get("infinitive markers"));
		short2long.put("num", Dict.get("numerals"));
		short2long.put("intj", Dict.get("interjections"));

		timing = false;
		timeout = false;
		timer = new Timer(1000);
		timer.addListener(this);
		timer.start();
	}
	/** Constructor, servername is the name of the server on which the 
	 * HitServer is running, language is the desired language and game is 
	 * the game
	 * @see HitServer
	 * @see #setLanguage(String)
	 * @see #setGame(String)
	 * @author Rasmus Lock Larsen */
	public HitClient(String servername, String language, String game) {
		this(servername, DEFAULT_PORT, language, game);
	}
	/** Checks to see, if the connection timed out! */
	public void tick() {
		if (timing) {
			time = System.currentTimeMillis() - start_time;
			if (time >= TIME_TIMEOUT) {
				try {
					close();
					socket.close();
				} catch (IOException x) {
				}
				timeout = true;
			}
		}
	}
	protected void startTiming() {
		start_time = System.currentTimeMillis();
		timing = true;
	}
	// ***********************************************************************
	// * HitServer calls
	// ***********************************************************************
	/** Reads a line from the connection, and resets the timing */
	protected String read() throws IOException {
		timing = false;
		start_time = System.currentTimeMillis();
		timing = true;
		String res = in.readLine();
		timing = false;
		return res;
	}
	/** Establishes a connection to the server */
	public void connect() throws IOException {
		timeout = false;
		socket = new Socket(servername, port);
		in = new DataInputStream(socket.getInputStream());
		out = new PrintStream(socket.getOutputStream());
	}
	/** Disconnects from the server (closes the connection) */
	public void close() throws IOException {
		out.println(DISCONNECT);
		timing = false;
	}
	/** Returns the HighScore denoted by the string s
	 * @see visl.rasmus.hitserver.server.HitServerInterface */
	public Vector getHighScore(String s) {
		Vector v = new Vector();
		try {
			connect();
			out.println(GET_HISCORE + language + ";" + game + ";" + s);
			String line = read();
			while (!line.equals(EOT)) {
				v.addElement(new HighScoreItem(line));
				line = read();
			}
			close();
		} catch (Exception e) {
			System.out.println(
				"Exception while communicating with the server:" + e);
		}
		return v;
	}
	/** Returns the overall position on the highscore given the score and
	 * hitrate (on the T100Y highscore) */
	public int getPosition(int score, float hitrate) {
		int r = -1;
		try {
			connect();
			out.println(
				QUE_HISCORE
					+ language
					+ ";"
					+ game
					+ ";"
					+ ((int) (score * 100 + hitrate)));
			r = Integer.parseInt(read());
			close();
		} catch (Exception e) {
			System.out.println(
				"Exception while retrieving highscore position:" + e);
		}
		return r;
	}
	/** Returns a vector with num random sentences that all contains at least
	 * one word with the form <CODE>form</CODE> and all has at least 
	 * <CODE>min</CODE> words and less than <CODE>max</CODE> words. */
	public Vector getRandomSentencesWithForm(
		String form,
		int min,
		int max,
		int num) {
		Vector res = new Vector();
		int ix = 0;
		try {
			connect();
			out.println(
				GET_SENTENCES
					+ language
					+ ";"
					+ game
					+ ";"
					+ form
					+ ";"
					+ min
					+ ";"
					+ max
					+ ";"
					+ num);
			String line = read();
			while (!line.startsWith(EOT)) {
				//System.out.println(line);
				res.addElement(new Sentence(line));
				line = read();
			}
			close();
		} catch (Exception e) {
			System.out.println("Exception while retreiving sentences:" + e);
			System.out.println("res = " + res + ", |res| = " + res.size());
			res = null;
		}
		return res;
	}
	/** Returns the <CODE>num</CODE> sentence in the current language 
	 * @see visl.rasmus.hitserver.data.Sentence */
	public Sentence getSentence(int num) {
		Sentence res = null;
		try {
			connect();
			out.println(GET_SENTENCE + language + ";" + game + ";" + num);
			String line = read();
			res = new Sentence(line);
			close();
		} catch (Exception e) {
			System.out.println("Exception while retreiving sentence:" + e);
		}
		return res;
	}
	/** Returns a random sentence from the current language
	 * @see visl.rasmus.hitserver.data.Sentence */
	public Sentence getRandomSentence() {
		Sentence res = null;
		try {
			connect();
			out.println(GET_SENTENCE + language + ";" + game + ";" + (-1));
			String line = read();
			res = new Sentence(line);
			close();
		} catch (Exception e) {
			System.out.println("Exception while retreiving sentence:" + e);
		}
		return res;
	}
	/** Returns a random sentence with min min words and max max words 
	 * @see visl.rasmus.hitserver.data.Sentence */
	public Sentence getRandomSentence(int min, int max) {
		Sentence res = null;
		try {
			connect();
			out.println(
				GET_SENTENCE + language + ";" + game + ";" + min + ";" + max);
			String line = read();
			res = new Sentence(line);
			close();
		} catch (Exception e) {
			System.out.println("Exception while retreiving sentence:" + e);
		}
		return res;
	}
	/** Returns a vector containing <CODE>num</CODE> random trees that satisfy
	 * the criteria that there is atleast <CODE>min</CODE> and no more 
	 * than <CODE>max</CODE> words in the tree (if less than <CODE>num</CODE>
	 * trees satisfies the criteria, the vector is truncated)
	 * @see visl.rasmus.hitserver.data.TreeNode */
	public Vector getRandomTrees(int min, int max, int num) {
		Vector res = new Vector();
		try {
			TreeNode nn = new TreeNode();
			connect();
			out.println(
				GET_TREES
					+ language
					+ ";"
					+ game
					+ ";"
					+ min
					+ ";"
					+ max
					+ ";"
					+ num);
			String line = read();
			while (!line.equals(EOT)) {
				//System.out.println("Got: " + line);
				TreeNode n = TreeNode.buildTree(line);
				res.addElement(n);
				line = read();
			}
			close();
		} catch (Exception x) {
			System.out.println("Exception while retreiving tree: " + x);
		}
		return res;
	}
	/** Adds to the highscore of the current language in the current game
	 * with the name <CODE>name</CODE> and the score <CODE>score</CODE> */
	public void putHighScore(
		String name,
		String country,
		int score,
		float hitrate) {
		HighScoreItem hsi =
			new HighScoreItem(
				name,
				country,
				score,
				new Date(System.currentTimeMillis()),
				hitrate);
		try {
			connect();
			out.println(PUT_HISCORE + language + ";" + game + ";" + hsi);
			close();
		} catch (Exception e) {
			System.out.println(
				"Exception while trying to add to HighScore:" + e);
		}
	}
	/** Returns a vector containing strings that denotes all the wordclasses
	 * in the current language. 
	 * @param wc can be "all", "minor", "major", "nv". */
	public Vector getWordClasses(String wc) {
		Vector res = new Vector();
		try {
			connect();
			out.println(GET_WCS + language + ";" + wc);
			String s = read();
			close();
			int ix = s.indexOf(':') + 1;
			int jx = s.indexOf(':', ix);
			while (ix < s.length() - 1) {
				res.addElement(s.substring(ix, jx));
				ix = jx + 1;
				jx = s.indexOf(':', ix);
			}
		} catch (Exception e) {
			System.out.println(
				"HitClient:getWordClasses(" + language + ":" + e + ")");
		}
		return res;
	}
	/** Returns a vector containing the functions
	 * @param desc determines whether the format should be "short" or "long" */
	public Vector getFunctions(String desc) {
		Vector res = new Vector();
		try {
			connect();
			out.println(GET_FUNCS + language + ";" + desc);
			String s = read();
			close();
			int ix = s.indexOf('|') + 1;
			int jx = s.indexOf('|', ix);
			while (ix < s.length() - 1) {
				res.addElement(s.substring(ix, jx));
				ix = jx + 1;
				jx = s.indexOf('|', ix);
			}
		} catch (Exception e) {
			System.out.println(
				"HitClient:getFunctions(" + language + ":" + e + ")");
		}
		return res;
	}
	/** Returns all the languages that are supported by the database. */
	public String[] getLanguages() {
		Vector v =
			sendQuery(
				SQL_COMMAND
					+ "2:q:SELECT language FROM "
					+ "settings ORDER BY language");
		String res[] = new String[v.size()];
		for (int i = 0; i < res.length; i++)
			res[i] = (String) v.elementAt(i);
		return res;
	}
	/** Returns all the languages that are supported by the current game. */
	public String[][] getLanguages4Game() {
		Vector v =
			sendQuery(
				SQL_COMMAND
					+ "2:q:SELECT language,name FROM "
					+ "settings WHERE LOCATE(':"
					+ game
					+ ":', games) != 0 "
					+ "ORDER BY language");
		String res[][] = null;
		if (v.size() > 0) {
			res = new String[v.size()][2];
			for (int i = 0; i < v.size(); i++) {
				String s = (String) v.elementAt(i);
				res[i][0] = s.substring(0, s.indexOf(" & "));
				res[i][1] = s.substring(s.indexOf(" & ") + 3);
			}
		}
		return res;
	}
	/** Returns all the possible games */
	public String[] getGames() {
		Vector v = sendQuery(SQL_COMMAND + "2:q:SELECT games FROM settings");

		Vector v2 = new Vector();
		for (int i = 0; i < v.size(); i++) {
			String val = (String) v.elementAt(i);
			String s = "";
			int ix = val.indexOf(":") + 1;
			int jx = val.indexOf(":", ix);
			while (jx != -1) {
				s = val.substring(ix, jx);
				if (!v2.contains(s))
					v2.addElement(s);
				ix = jx + 1;
				jx = val.indexOf(":", ix);
			}
		}
		String res[] = new String[v2.size()];
		for (int i = 0; i < v2.size(); i++)
			res[i] = (String) v2.elementAt(i);
		return res;
	}
	/** Returns the encoding of the current language. NB: If none is present 
	 * in the database ISO8859_1 is returned */
	public String getEncoding() {
		Vector v =
			sendQuery(
				SQL_COMMAND
					+ "2:q:SELECT encoding FROM settings "
					+ "WHERE language='"
					+ language
					+ "'");
		if (v.size() == 1)
			return (String) v.elementAt(0);
		return "ISO8859_1";
	}
	/** Sends the string <CODE>command</CODE> as a command to the HitServer */
	protected void sendCommand(String command) {
		try {
			connect();
			out.println(command);
			close();
		} catch (Exception x) {
			System.out.println(
				"Exception while trying to execute command ["
					+ command
					+ "]: "
					+ x);
			x.printStackTrace();
		}
	}
	/** Sends a query-command to the HitServer. Returns a vector
	 * containing all the lines that are returned. */
	protected Vector sendQuery(String command) {
		Vector res = new Vector();
		try {
			connect();
			out.println(command);
			String line = read();
			while (!line.equals(EOT)) {
				res.addElement(line);
				line = read();
			}
			close();
		} catch (Exception x) {
			System.out.println(
				"Exception while trying to execute command ["
					+ command
					+ "]: "
					+ x);
		}
		return res;
	}
	// ***********************************************************************
	// * Field Methods
	// ***********************************************************************
	/** Changes the language used */
	public void setLanguage(String lang) {
		language = lang;
	}
	/** Changes the game parameter */
	public void setGame(String gam) {
		game = gam;
	}
	public boolean isTimeOut() {
		return timeout;
	}
}
