package visl.rasmus.hitserver.data;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;

/**
* A node in the syntax tree.
* @author Martin Carlsen
* @author Rasmus Lock Larsen, VISL 5th of april 2001 - Formerly known as 
* SyntaxModel in VISLApplet
*/
public class TreeNode {
    protected long id;
    private String function;	// The function of the node. Or the lexical 
                                //  item if the node is a leaf node.
    private String form;	// The form of the node.
    private String morphology;	// The morphology of the node.
    private int depth;		// The depth of the node.
    private int disc;           // Disconinuities. Whether this node 
                                //  continues to the right and/or left
    private TreeNode leftDisc, rightDisc; // left and right nodes of 
                                          //  the discontinuity (if any);
    /** Parent of the node. */
    protected TreeNode parent;
    /** The first child of the node. */
    protected TreeNode firstChild;
    /** The sibling next to the node. */
    protected TreeNode nextSibling;
    /** The number of children of the node. */
    protected int childCount;

    // Discontinuity constants
    public static final int NONE = 0;
    public static final int LEFT = 1;
    public static final int RIGHT = 2;
    public static final int BOTH = 3;
	
    /**
     * Constructor: Builds the tree.
     * @param nodes An array of strings defining the nodes and structure of 
     * the tree.
     * Each string must be of the following format:
     * A number of '='s indicating the depth of the node in the tree followed 
     * by the function followed by ':' followed by the form optionally 
     * followed by the morphology in parentheses.
     * <CODE>\=*{function}\:{form}[\({morphology}\)]?</CODE>
     * See the parameter parsing HOWTO.
     * @param nodeDef The applet parameter that defines this node.
     */
    public TreeNode(String nodeDef) {
	this.id = -1;
	setNodeDef(nodeDef);
    }
    public TreeNode() {
	firstChild = null;
	parent = null;
	childCount = 0;
	nextSibling = null;
	disc = NONE;
	function = "";
	form = "";
    }
    public TreeNode(TreeNode n) {
	firstChild = n.firstChild;
	parent = n.parent;
	childCount = n.childCount;
	nextSibling = n.nextSibling;
	disc = n.disc;
	function = n.function;
	form = n.form;
    }
    public TreeNode(String nodeDef, long id){
	this.id = id;
	setNodeDef(nodeDef);
    }
	
    private void setNodeDef(String nodeDef){
	int from,to;
		
	// Extract depth.
	depth = nodeDef.lastIndexOf('=')+1;
	from = depth;
		
	// Extract left-discontinuity.
	disc = NONE;
	if(nodeDef.charAt(from)=='-'){
	    disc += LEFT;
	    from++;
	}
		
	// Extract function
	if(from == nodeDef.length()){ // empty node
	    function = "";
	    form = "";
	    morphology = "";
	    return;
	} else {
	    to = nodeDef.indexOf(':');
	    if (to < 0) { // Node is a leaf node.
		to = nodeDef.length();
		function = nodeDef.substring(from,to);
		form = "";
		morphology = "";
		return;
	    }
	    function = nodeDef.substring(from,to);
	}
		
	// Extract form.
	from = nodeDef.indexOf(':')+1;
	if(nodeDef.lastIndexOf('(') < from){ // No information about morphology.
	    to = nodeDef.indexOf('/', from);
	    if (to == -1)
		to = nodeDef.length()-1;
	}
	else{
	    to = nodeDef.indexOf('/', from);
	    if (to == -1)
		to = nodeDef.lastIndexOf('(')-1;
	}
		
	// Extract right-discontinuity here.
	if(nodeDef.charAt(to) == '-'){
	    disc += RIGHT;
	    to--;
	}
	form = nodeDef.substring(from,to+1);
		
	// Extract morphology.
	if(nodeDef.lastIndexOf('(') < 0){ // No information about morphology
	    morphology = "";
	}else{
	    from = nodeDef.lastIndexOf('(')+1;
	    to = nodeDef.lastIndexOf(')')-1;
	    morphology = nodeDef.substring(from,to+1);
	}
    }
	
    /**
     * Connect discontinous parts.
     * This is invoked from the Builder of the tree, each time a subtree 
     * has been finished.
     */
    //
    // This method conceptually belongs in SyntaxModelBuilder,
    // but encapsulation (of attributes related to discontinuities) will 
    // be broken then.
    // There must be a better way.
    //
    public void doDiscontinuities()
    {
	// search for discontinuities
	for(TreeNode node = (TreeNode)(getFirstChild()); 
	    node != null; 
	    node = (TreeNode)(node.getNextSibling())) 
	    {
		if(node.continuesRight()) {
		    // search to the right for the next part.
		    for(TreeNode n2 = (TreeNode)(node.getNextSibling()); 
			n2 != null;
			n2 = (TreeNode)(n2.getNextSibling()))
			{
			    if (n2.continuesLeft() & 
				n2.getFunction().equals(node.getFunction()) )
				{
				    node.rightDisc = n2;
				    n2.leftDisc = node;
				    break;
				}
			}
		}
	    }
    }
	
    public TreeNode getLeftmost() { 
	return (leftDisc==null) ? this:leftDisc.getLeftmost(); 
    }
    public TreeNode getRightmost() { 
	return (rightDisc==null) ? this : rightDisc.getRightmost(); 
    }
    public TreeNode leftDiscontinuity(){ return leftDisc; }
    public TreeNode rightDiscontinuity(){ return rightDisc; }
    public boolean continuesLeft(){ return (disc&LEFT) == LEFT; }
    public boolean continuesRight(){ return (disc&RIGHT) == RIGHT; }
    public boolean isDiscontinous(){ return disc != NONE; }
    public String getFunction(){ return function; }
    public String getForm(){ return form; }
    public String getMorphology(){ return morphology; }
    public int getDepth(){ return depth; }
    public void setFunction(String f){ function = f; }
    public void setForm(String f){ form = f; }
    public void setMorphology(String m){ morphology = m; }
    public void setDepth(int d){ depth = d; }

    // ********************************************************************
    // * FORMER TREEMODEL
    // ********************************************************************

    /** Get whether the node is a leaf node.
     *@return Whether the node is a leaf node. */
    public boolean     isLeaf()           { return firstChild == null; }

    /** Get whether the node is a root node.
     *@return Whether the node is a root node. */
    public boolean     isRoot()           { return parent == null; }
    
    /** Get the parent of the node.
     * @return The parent of the node. <TT>null</TT> if the node is a
     * root node. */
    public TreeNode   getParent()        { return parent; }
    
    /** Get the children of the node.
     *@return An <TT>Enumeration</TT> of the node's children. */
    public Enumeration getChildren()      { return new ChildEnum(firstChild); }
    
    /** Get the number of children.
     *@return The number of children of the node. */
    public int         getChildCount()    { return childCount; }
    
    /** Get the root node of the tree.
     *@return The root node of the tree. */
    public TreeNode   getRoot()          { return parent==null ? this : parent.getRoot(); }
    
    /** Get the last sibling in the list of siblings. 
     *@return The last of the node's siblings. */
    public TreeNode   getLastSibling()   { return nextSibling==null ? this : nextSibling.getLastSibling(); }
    
    /** Get the first child of the node. 
     *@return The first of the node's children. */
    public TreeNode   getFirstChild()    { return firstChild; }
    
    /** Get the next sibling in the list of siblings.
     *@return The next of the node's sibling. */
    public TreeNode   getNextSibling()   { return nextSibling; }
    /** Get the previous sibling in the list of siblings.
     * @return the previous of the node's siblings.
     */
    public TreeNode getPreviousSibling() {
	TreeNode n = (parent != null ? parent.firstChild : null);
	if (n == this || n == null) return null;
	while (n.nextSibling != null &&
	       n.nextSibling != this) n = n.nextSibling;
	return n;
    }
    
    /**
     * Insert new child node (along with its children) into model 
     * as the first child of this node.
     * Removes node from the tree it was previously in.
     *@param n Node to insert.
     */
    public void insertChild(TreeNode n){
	n.remove();
	n.parent = this;
	n.nextSibling = firstChild;
	firstChild = n;
	childCount++;
    }
    
    /**
     * Append new child node (along with its children) into model 
     * as the last child of this node.
     * Removes node from the tree it was previously in.
     *@param n Node to insert.
     */
    public void appendChild(TreeNode n){
	if(firstChild == null) insertChild(n);
	else firstChild.getLastSibling().insertSibling(n);
    }
    
    /**
     * Insert new sibling node (along with its children) into model.
     * Removes node from the tree it was previously in.
     *@param n Node to insert.
     */
    public void insertSibling(TreeNode n){
	n.remove();
	n.parent = parent;
	n.nextSibling = nextSibling;
	nextSibling = n;
	if(parent!=null) parent.childCount++;
    }
    
    /**
     * Remove node (along with its children) from model.
     */
    public void remove(){
	if(parent==null) return;
	if(parent.firstChild == this)
	    parent.firstChild = nextSibling;
	else
	    for(TreeNode n = parent.firstChild; n != null; n = n.nextSibling)
		if(n.nextSibling == this){
		    n.nextSibling = nextSibling;
		    break;
		}
	parent.childCount--;
    }
    
    /**@see  #execute(Visitor, int) */
    public static final int DEPTH_FIRST     = 0x00000001;
    /**@see  #execute(Visitor, int) */
    public static final int BREADTH_FIRST   = 0x00000010;
    /**@see  #execute(Visitor, int) */
    public static final int UNBOUND         = 0x00000100;
    /**@see  #execute(Visitor, int) */
    public static final int BOUND           = 0x00001000;
    /**@see  #execute(Visitor, int) */
    public static final int CHILDREN_FIRST  = 0x00010000;
    /**@see  #execute(Visitor, int) */
    public static final int SIBLINGS_FIRST  = 0x00100000;
    /**@see  #execute(Visitor, int) */
    public static final int TO_ROOT         = 0x01000000;
    /**@see  #execute(Visitor, int) */
    public static final int FROM_ROOT       = 0x10000000;
    /** Default strategy (DEPTH_FIRST)
     *@see #execute(Visitor, int) */
    public static final int DEFAULT         = DEPTH_FIRST;
    
    /**
     * Execution of command on tree.
     * The possible execution order strategies are:
     * <PRE>this, children, siblings : DEPTH_FIRST
     * children, this, siblings : DEPTH_FIRST + CHILDREN_FIRST
     * children, siblings, this : DEPTH_FIRST + CHILDREN_FIRST + SIBLINGS_FIRST
     * this, siblings, children : BREADTH_FIRST
     * siblings, this, children : BREADTH_FIRST + SIBLINGS_FIRST
     * siblings, children, this : BREADTH_FIRST+SIBLINGS_FIRST+CHILDREN_FIRST
     * this, parent, ... , root : TO_ROOT
     * root, ... , parent, this : FROM_ROOT<PRE>
     * Strategies other than <TT>TO_ROOT</TT> and <TT>FROM_ROOT</TT> may 
     * be <TT>BOUND</TT>
     * (e.g. BOUND + DEPTH_FIRST),
     * indicating that <TT>goWider()</TT> and <TT>goDeeper()</TT> is used to 
     * test which nodes
     * to execute the command for, or they may be <TT>UNBOUND</TT> indicating
     * the command is executed
     * for all nodes. <TT>UNBOUND</TT> is default and need not be specified.
     * <P>Any other strategy than the strategies indicated above is an invalid
     * strategy.
     * (e.g. BREADTH_FIRST + CHILDREN_FIRST)
     * @param cmd Command to execute.
     * @param strategy The strategy that specifies the order of execution.
     * @exception IllegalArgumentException If the specified strategy is invalid
     */
    public void execute(Visitor cmd, int strategy){
	boolean unbound = (strategy & BOUND) == 0;
	int s = strategy & 0x11110011; // ignore BOUND and UNBOUND bits.
	switch(s){
	case TO_ROOT:
	    cmd.execute(this);
	    if(parent != null) parent.execute(cmd, strategy);
	    break;
	case FROM_ROOT:
	    if(parent != null) parent.execute(cmd, strategy);
	    cmd.execute(this);
	    break;
	case DEPTH_FIRST:
	    cmd.execute(this);
	    if((unbound || cmd.goDeeper(this)) && firstChild != null) 
		firstChild.execute(cmd, strategy);
	    if((unbound || cmd.goWider(this)) && nextSibling != null) 
		nextSibling.execute(cmd, strategy);
	    break;
	case DEPTH_FIRST | CHILDREN_FIRST:
	    if((unbound || cmd.goDeeper(this)) && firstChild != null) 
		firstChild.execute(cmd, strategy);
	    cmd.execute(this);
	    if((unbound || cmd.goWider(this)) && nextSibling != null) 
		nextSibling.execute(cmd, strategy);
	    break;
	case DEPTH_FIRST | CHILDREN_FIRST | SIBLINGS_FIRST:
	    if((unbound || cmd.goDeeper(this)) && firstChild != null) 
		firstChild.execute(cmd, strategy);
	    if((unbound || cmd.goWider(this)) && nextSibling != null) 
		nextSibling.execute(cmd, strategy);
	    cmd.execute(this);
	    break;
	case BREADTH_FIRST:
	    cmd.execute(this);
	    if((unbound || cmd.goWider(this)) && nextSibling != null) 
		nextSibling.execute(cmd, strategy);
	    if((unbound || cmd.goDeeper(this)) && firstChild != null) 
		firstChild.execute(cmd, strategy);
	    break;
	case BREADTH_FIRST | SIBLINGS_FIRST:
	    if((unbound || cmd.goWider(this)) && nextSibling != null) 
		nextSibling.execute(cmd, strategy);
	    cmd.execute(this);
	    if((unbound || cmd.goDeeper(this)) && firstChild != null) 
		firstChild.execute(cmd, strategy);
	    break;
	case BREADTH_FIRST | SIBLINGS_FIRST | CHILDREN_FIRST:
	    if((unbound || cmd.goWider(this)) && nextSibling != null) 
		nextSibling.execute(cmd, strategy);
	    if((unbound || cmd.goDeeper(this)) && firstChild != null) 
		firstChild.execute(cmd, strategy);
	    cmd.execute(this);
	    break;
	default:
	    throw new IllegalArgumentException("Invalid strategy.");
	}
    }
  
    /** Enumeration of TreeNode siblings. */
    protected class ChildEnum implements Enumeration{
	TreeNode node;
    
	public ChildEnum()           { node = null; }
	public ChildEnum(TreeNode n){ node = n; }
    
	public boolean hasMoreElements(){ return node != null; }
	public Object nextElement() throws NoSuchElementException{
	    if(node==null) throw new NoSuchElementException();
	    TreeNode n = node;
	    node = node.nextSibling;
	    return n;
	}
    }
    public String toString() {
	return toString(this);
    }
    public boolean equals(Object o) {
	if (!(o instanceof TreeNode)) return false;
	return toString().equals(((TreeNode)o).toString());
    }
    /** Creates a string representation of the subtree n */
    static public String toString(TreeNode n) {
	return "Tree["+ toString(n, n.id) +"]";
    }
    /** Creates a string representing this node, and it descendents 
     * Syntax: {pix:nsix:fchix:ldix:rdix:func:form:morp:disc} 
     */
    static protected String toString(TreeNode n, long pid) {
	if (n == null) return "";
	String res = "{"+ n.form.trim() + "|" + n.function + "|" +
	    (n.parent != null ? n.parent.id-pid : -1) + "|" +
	    (n.nextSibling != null ? n.nextSibling.id-pid : -1) + "|" +
	    (n.firstChild != null ? n.firstChild.id-pid : -1) + "|" +
	    (n.leftDisc != null ? n.leftDisc.id-pid : -1) + "|" +
	    (n.rightDisc != null ? n.rightDisc.id-pid : -1) + "|" + 
	    n.morphology + "|" + 
	    n.disc + "}";
	res += toString(n.firstChild, pid);
	res += toString(n.nextSibling, pid);
	return res;
    }
    static public TreeNode buildTree(String s) {
	//System.out.println("TreeNode:buildTree("+s+")");
	int ix = 0, jx = 0;
	TreeNode res = null;
	Vector nodes = new Vector();
	try {
	    ix = s.indexOf("[");
	    if (s.substring(0,ix).equals("Tree")) {
		jx = ix + 1;
		while (ix != -1 && jx != -1) {
		    ix = s.indexOf('{', jx);
		    jx = s.indexOf('}', ix);
		    if (jx != -1 && ix != -1) {
			nodes.addElement(s.substring(ix+1, jx));
		    }
		}
	    } else throw new RuntimeException("Not the correct classname ["+s+"]");
	    TreeNode tnds[] = new TreeNode[nodes.size()];
	    for (int i = 0; i < tnds.length; i++) {
		tnds[i] = new TreeNode();
	    }
	    for (int i = 0; i < tnds.length; i++) {
		String n = (String)nodes.elementAt(i);
		jx = -1;
		// NOTE: Something might go wrong here!
		jx = n.indexOf('|', ix=jx+1);
		tnds[i].form = n.substring(ix, jx);

		// NOTE: Something might go wrong here!
		jx = n.indexOf('|', ix=jx+1);
		tnds[i].function = n.substring(ix, jx);

		jx = n.indexOf('|', ix=jx+1);

		int num = Integer.parseInt(n.substring(ix, jx));
		tnds[i].parent = (num == -1 ? null : tnds[num]);

		jx = n.indexOf('|', ix=jx+1);
		num = Integer.parseInt(n.substring(ix, jx));
		tnds[i].nextSibling = (num == -1 ? null : tnds[num]);

		jx = n.indexOf('|', ix=jx+1);
		num = Integer.parseInt(n.substring(ix, jx));
		tnds[i].firstChild = (num == -1 ? null : tnds[num]);

		jx = n.indexOf('|', ix=jx+1);
		num = Integer.parseInt(n.substring(ix, jx));
		tnds[i].leftDisc = (num == -1 ? null : tnds[num]);

		jx = n.indexOf('|', ix=jx+1);
		num = Integer.parseInt(n.substring(ix, jx));
		tnds[i].rightDisc = (num == -1 ? null : tnds[num]);
		
		jx = n.indexOf('|', ix=jx+1);
		tnds[i].morphology = n.substring(ix, jx);

		ix = jx + 1;
		jx = n.length();
		num = Integer.parseInt(n.substring(ix, jx));
		tnds[i].disc = num;
	    }
	    // Do the depths
	    for (int i = 0; i < tnds.length; i++) {
		if (tnds[i].parent != null) 
		    tnds[i].depth = tnds[i].parent.depth + 1;
		tnds[i].childCount = 0;
		TreeNode tmpnode = tnds[i].firstChild;
		while (tmpnode != null) {
		    tnds[i].childCount++;
		    tmpnode = tmpnode.nextSibling;
		}
		    
	    }
	    res = tnds[0];
	} catch (Exception x) {
	    System.out.println("TreeNode:Exception when resolving data: " + x);
	    System.out.println("["+s+"], ix = " + ix + ", jx = " + jx);
	}
	return res;
    }
    /** Returns the number of leafes on this tree */
    public int leafCount() {
	// Do some mambo to avoid the rest of the tree!
	TreeNode p = parent, n = nextSibling;
	parent = null;
	nextSibling = null;
	LeafCounter lc = new LeafCounter();
	this.execute(lc, DEFAULT);
	parent = p;
	nextSibling = n;
	return lc.getCount();
    }
    /** Returns the maximum depth of the branches in this tree */
    public int maxDepth() {
	// Do some mambo to avoid the rest of the tree!
	TreeNode p = parent, n = nextSibling;
	parent = null;
	nextSibling = null;
	DepthCounter dc = new DepthCounter();
	this.execute(dc, DEFAULT);
	parent = p;
	nextSibling = n;
	return dc.getCount();
    }
    public boolean isDiscontinousTree() {
	// Do some mambo to avoid the rest of the tree!
	TreeNode p = parent, n = nextSibling;
	parent = null;
	nextSibling = null;
	DiscChecker dc = new DiscChecker();
	this.execute(dc, DEFAULT);
	parent = p;
	nextSibling = n;
	return dc.isDiscontinous();
    }
     /** Returns the sentence formed by the leafs of the tree */
    public String getString() {
	// Do some mambo to avoid the rest of the tree!
	TreeNode p = parent, n = nextSibling;
	parent = null;
	nextSibling = null;
	StringFormer sf = new StringFormer();
	this.execute(sf, DEFAULT);
	parent = p;
	nextSibling = n;
	return sf.getString();
    }
    /** Returns a vector containing all the "words" (leafs) in the tree 
     * @see visl.rasmus.hitserver.data.Word */
    public Vector getWords() {
	// Do some mambo to avoid the rest of the tree!
	TreeNode p = parent, n = nextSibling;
	parent = null;
	nextSibling = null;
	WordFormer wf = new WordFormer();
	this.execute(wf, DEFAULT);
	parent = p;
	nextSibling = n;
	return wf.getWords();
    }
    /** Returns a vector containing all the "words" (leafs) in the tree.
     * The words also contain morphological information.
     * @see visl.rasmus.hitserver.data.Word */
    public Vector getMorphWords() {
	// Do some mambo to avoid the rest of the tree!
	TreeNode p = parent, n = nextSibling;
	parent = null;
	nextSibling = null;
	MorphologyFormer mf = new MorphologyFormer();
	this.execute(mf, DEFAULT);
	parent = p;
	nextSibling = n;
	return mf.getMorphWords();
    }
    public void filter() {
	FilterCommand fc = new FilterCommand();
	this.execute(fc, DEFAULT);
    }
    public Vector getFormFunction() {
	FormFunctionCommand ffc = new FormFunctionCommand();
	this.execute(ffc, DEFAULT);
	return ffc.getResult();
    }
    // ********************************************************************
    // * COMMAND CLASSES
    // ********************************************************************
    static protected class DepthCounter extends UnboundCommand {
	int count;
	public DepthCounter() {
	    count = 0;
	}
	public void execute(TreeNode n) {
	    if (n.isLeaf())
		if (n.depth >= count) count = n.depth;
	}
	public int getCount() { return count; }
    }
    static protected class LeafCounter extends UnboundCommand {
	int count;
	public LeafCounter() {
	    count = 0;
	}
	public void execute(TreeNode n) {
	    if (n.isLeaf()) count++;
	}
	public int getCount() { return count; }
    }
    static protected class DiscChecker extends UnboundCommand {
	boolean disc;
	public DiscChecker() { disc = false; }
	public void execute(TreeNode n) {
	    if (n.isDiscontinous()) disc = true;
	}
	public boolean isDiscontinous() { return disc; }
    }
    static protected class StringFormer extends UnboundCommand {
	StringBuffer sb;
	public StringFormer() { sb = new StringBuffer(); }
	public void execute(TreeNode n) {
	    if (n.isLeaf()) {
		sb.append(n.function+" ");
	    }
	}
	public String getString() { 
	    return sb.toString().substring(0,sb.length()-1); 
	}
    }
    static protected class WordFormer extends UnboundCommand {
	Vector v;
	public WordFormer() { v = new Vector(); }
	public void execute(TreeNode n) {
	    if (n.isLeaf()) {
		Word w = new Word(n.function, n.parent.form);
		v.addElement(w);
	    }
	}
	public Vector getWords() { 
	    return v;
	}
    }
    static protected class MorphologyFormer extends UnboundCommand {
	Vector v;
	public MorphologyFormer() { v = new Vector(); }
	public void execute(TreeNode n) {
	    if (n.isLeaf()) {
		MorphWord w = new MorphWord(n.function, n.parent.form, n.parent.morphology);
		v.addElement(w);
	    }
	}
	public Vector getMorphWords() { 
	    return v;
	}
    }
    static protected class FilterCommand extends UnboundCommand {
	public void execute(TreeNode n) {
	    n.form = TreeDB.filterForm(n.form);
	    n.function = TreeDB.filterFunction(n.function);
	}
    }
    static protected class FormFunctionCommand extends UnboundCommand {
	protected Vector elems;
	public FormFunctionCommand() { elems = new Vector(); }
	public void execute(TreeNode n) {
	    if (!n.isLeaf()) 
		elems.addElement("F:"+n.function);
	    elems.addElement("f:"+n.form);
	}
	public Vector getResult() { return elems; }
    }
}

