package visl.rasmus.utils;

import java.awt.Event;
import java.awt.Frame;
import java.awt.Label;

import visl.resources.Dict;

/**
* This class provides loading of a number of classes in a separate Thread, 
* while displaying a simple progress meter.
*
* The easy way to use it is:
* <CODE>(new Thread(new ClassLoader( listOfClassNames ) ) ).start();</CODE>
* 
* A non-threaded way is:
* <CODE>(new ClassLoader( listOfClassNames ) ).run();</CODE>
*
* A non-graphical, non-threaded way is:
* <CODE>Loader.load( listOfClassNames );</CODE>
*
* @author Martin Carlsen (VISL)
*/
public class Loader extends Frame implements Runnable {
    /** The thread that is doing the loading of classes */
    private Thread loaderThread;
    /** The classes to load. */
    private String[] classes;
    
    /**
     * Default constructor.
     */
    private Loader(){
	super(Dict.get("Loading. Please Wait..."));
	resize(400,150);
    }
    
    /**
     *@param classes List of fully qualified names of classes to load
     */
    public Loader(String[] classes) {
	this();
	this.classes = classes;
    }
    
    /**
     *@param classes List of fully qualified names of classes to load
     */
    public void setClasses(String[] classes) {
	this.classes = classes;
    }
    
    /**
     * Load classes specified in constructor or setClasses();
     * This method implements the Runnable interface.
     */
    public void run() {
	loaderThread = Thread.currentThread();
	
	show();
	
	// Load classes while displaying progress.
	try {
	    removeAll();
	    add("Center", new Label(Dict.get("Please wait while the applet is loading.") ) );
	    String loading = Dict.get("Loaded")+" ";
	    Label l = new Label(loading + "   " + Dict.get("of") + "    ");
	    add("South",l);
	    validate();
	    
	    int i=0;
	    l.setText(loading + (i) + " "+Dict.get("of") + " " + classes.length);
	    for(i=0; i<classes.length; i++){
		Class c = Class.forName(classes[i]);
		l.setText(loading + (i+1) + " " + Dict.get("of") + " " + classes.length);
	    }
	}
	// Must catch exceptions from loadClasses
	catch(Throwable e){ // Display an error message and stop the applet.
	    stop(Dict.get("Applet stopped: Error while loading classes:"), e);
	    return;
	}
	dispose();
    }
	
    /**
     * Load classes.
     *@param classes List of fully qualified names of classes to load
     */
    public static void load(String[] classes) throws ClassNotFoundException {
	for(int i=0; i<classes.length; i++){
	    Class c = Class.forName(classes[i]);
	}
    }
    
    /**
     * Handle WINDOW_CLOSE event.
     */
    public boolean handleEvent(Event e){
	if(e.id == Event.WINDOW_DESTROY){
	    if (loaderThread != null) loaderThread.stop();
	    dispose();
	}
	return super.handleEvent(e);
    }
    
    /** Finalizer that disposes of the window. */
    public void finalize() throws Throwable
    {
	dispose();
	super.finalize();
    }
    
    /** Show an error message in the window, then stop loading.*/
    private void stop(String msg, Throwable x) {
	removeAll();
	add("Center",new Label(msg));
	add("South",new Label(x.toString()));
	if(loaderThread!=null) loaderThread.stop();
    }
}
